@@ -1,19 +1,24 @@
-pub use crate::core::indexed_impls::{PlaceDomain, PlaceIndex, PlaceSet};
+#![allow(dead_code)]
+
 use anyhow::{bail, Result};
-use log::warn;
-use rustc_data_structures::fx::FxHashMap as HashMap;
+use log::{trace, warn};
+use rustc_borrowck::consumers::BodyWithBorrowckFacts;
+use rustc_data_structures::fx::{FxHashMap as HashMap, FxHashSet as HashSet};
 use rustc_graphviz as dot;
+use rustc_hir::{def_id::DefId, BodyId};
 use rustc_middle::{
   mir::{
-    visit::{PlaceContext, Visitor},
+    pretty::write_mir_fn,
+    visit::{NonUseContext, PlaceContext, Visitor},
     *,
   },
-  ty::{self, RegionKind, RegionVid, Ty, TyCtxt, TyKind, TyS, TypeFoldable, TypeVisitor},
-};
-use rustc_mir::{
-  dataflow::{fmt::DebugWithContext, graphviz, Analysis, Results},
-  util::write_mir_fn,
+  ty::{
+    self, RegionKind, RegionVid, Ty, TyCtxt, TyKind, TyS, TypeFoldable, TypeVisitor,
+    WithOptConstParam,
+  },
 };
+use rustc_mir_dataflow::{fmt::DebugWithContext, graphviz, Analysis, Results};
+use rustc_mir_transform::MirPass;
 use rustc_span::Span;
 use rustc_target::abi::VariantIdx;
 use std::{
@@ -36,13 +41,20 @@ pub fn local_to_place(local: Local, tcx: TyCtxt<'tcx>) -> Place<'tcx> {
 
 struct CollectRegions<'tcx> {
   tcx: TyCtxt<'tcx>,
+  def_id: DefId,
   local: Local,
   place_stack: Vec<PlaceElem<'tcx>>,
   ty_stack: Vec<Ty<'tcx>>,
-  regions: HashMap<RegionVid, (Place<'tcx>, Mutability)>,
+  places: Option<HashSet<Place<'tcx>>>,
+  types: Option<HashSet<Ty<'tcx>>>,
+  regions: HashMap<RegionVid, Vec<(Place<'tcx>, Mutability)>>,
 }
 
 impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
+  fn tcx_for_anon_const_substs(&self) -> Option<TyCtxt<'tcx>> {
+    Some(self.tcx)
+  }
+
   fn visit_ty(&mut self, ty: Ty<'tcx>) -> ControlFlow<Self::BreakTy> {
     if self
       .ty_stack
@@ -52,9 +64,17 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
       return ControlFlow::Continue(());
     }
 
+    trace!("exploring local {:?} with {:?}", self.local, ty);
+
     self.ty_stack.push(ty);
 
     match ty.kind() {
+      _ if ty.is_box() => {
+        self.place_stack.push(ProjectionElem::Deref);
+        self.visit_ty(ty.boxed_ty());
+        self.place_stack.pop();
+      }
+
       TyKind::Tuple(fields) => {
         for (i, field) in fields.iter().enumerate() {
           self.place_stack.push(ProjectionElem::Field(
@@ -69,6 +89,10 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
       TyKind::Adt(adt_def, subst) => match adt_def.adt_kind() {
         ty::AdtKind::Struct => {
           for (i, field) in adt_def.all_fields().enumerate() {
+            if !field.vis.is_accessible_from(self.def_id, self.tcx) {
+              continue;
+            }
+
             let ty = field.ty(self.tcx, subst);
             self
               .place_stack
@@ -136,11 +160,23 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
       }
     };
 
+    if let Some(places) = self.places.as_mut() {
+      places.insert(Place {
+        local: self.local,
+        projection: self.tcx.intern_place_elems(&self.place_stack),
+      });
+    }
+
+    if let Some(types) = self.types.as_mut() {
+      types.insert(ty);
+    }
+
     self.ty_stack.pop();
     ControlFlow::Continue(())
   }
 
   fn visit_region(&mut self, region: ty::Region<'tcx>) -> ControlFlow<Self::BreakTy> {
+    trace!("visiting region {:?}", region);
     match region {
       RegionKind::ReVar(region) => {
         let mutability = if self
@@ -157,7 +193,11 @@ impl TypeVisitor<'tcx> for CollectRegions<'tcx> {
           local: self.local,
           projection: self.tcx.intern_place_elems(&self.place_stack),
         };
-        self.regions.insert(*region, (place, mutability));
+        self
+          .regions
+          .entry(*region)
+          .or_default()
+          .push((place, mutability));
       }
       RegionKind::ReStatic => {}
       _ => unreachable!("{:?}: {:?}", self.ty_stack.first().unwrap(), region),
@@ -171,19 +211,98 @@ pub fn interior_pointers<'tcx>(
   place: Place<'tcx>,
   tcx: TyCtxt<'tcx>,
   body: &Body<'tcx>,
-) -> HashMap<RegionVid, (Place<'tcx>, Mutability)> {
+  def_id: DefId,
+) -> HashMap<RegionVid, Vec<(Place<'tcx>, Mutability)>> {
   let ty = place.ty(body.local_decls(), tcx).ty;
   let mut region_collector = CollectRegions {
     tcx,
+    def_id,
     local: place.local,
-    place_stack: vec![],
+    place_stack: place.projection.to_vec(),
     ty_stack: Vec::new(),
     regions: HashMap::default(),
+    places: None,
+    types: None,
   };
   region_collector.visit_ty(ty);
   region_collector.regions
 }
 
+pub fn interior_places<'tcx>(
+  place: Place<'tcx>,
+  tcx: TyCtxt<'tcx>,
+  body: &Body<'tcx>,
+  def_id: DefId,
+) -> Vec<Place<'tcx>> {
+  let ty = place.ty(body.local_decls(), tcx).ty;
+  let mut region_collector = CollectRegions {
+    tcx,
+    def_id,
+    local: place.local,
+    place_stack: place.projection.to_vec(),
+    ty_stack: Vec::new(),
+    regions: HashMap::default(),
+    places: Some(HashSet::default()),
+    types: None,
+  };
+  region_collector.visit_ty(ty);
+  region_collector.places.unwrap().into_iter().collect()
+}
+
+pub fn interior_types<'tcx>(
+  place: Place<'tcx>,
+  tcx: TyCtxt<'tcx>,
+  body: &Body<'tcx>,
+  def_id: DefId,
+) -> Vec<Ty<'tcx>> {
+  let ty = place.ty(body.local_decls(), tcx).ty;
+  let mut region_collector = CollectRegions {
+    tcx,
+    def_id,
+    local: place.local,
+    place_stack: place.projection.to_vec(),
+    ty_stack: Vec::new(),
+    regions: HashMap::default(),
+    places: None,
+    types: Some(HashSet::default()),
+  };
+  region_collector.visit_ty(ty);
+  region_collector.types.unwrap().into_iter().collect()
+}
+
+pub fn arg_mut_ptrs<'tcx>(
+  args: &[Place<'tcx>],
+  tcx: TyCtxt<'tcx>,
+  body: &Body<'tcx>,
+  def_id: DefId,
+) -> Vec<Place<'tcx>> {
+  args
+    .iter()
+    .map(|place| {
+      interior_pointers(*place, tcx, body, def_id)
+        .into_iter()
+        .map(|(_, places)| {
+          places
+            .into_iter()
+            .filter_map(|(place, mutability)| match mutability {
+              Mutability::Mut => Some(place),
+              Mutability::Not => None,
+            })
+        })
+        .flatten()
+        .map(|place| tcx.mk_place_deref(place))
+    })
+    .flatten()
+    .collect::<Vec<_>>()
+}
+
+pub fn arg_places<'tcx>(args: &[Operand<'tcx>]) -> Vec<Place<'tcx>> {
+  args
+    .iter()
+    .filter_map(|arg| operand_to_place(arg))
+    .collect::<Vec<_>>()
+}
+
 pub fn hashmap_merge<K: Eq + Hash, V>(
   mut h1: HashMap<K, V>,
   h2: HashMap<K, V>,
@@ -239,6 +358,18 @@ impl PlaceRelation {
       });
 
     let is_sub_part = part_place.projection.len() >= whole_place.projection.len();
+    let remaining_projection = if is_sub_part {
+      &part_place.projection[whole_place.projection.len()..]
+    } else {
+      &whole_place.projection[part_place.projection.len()..]
+    };
+
+    if remaining_projection
+      .iter()
+      .any(|elem| matches!(elem, ProjectionElem::Deref))
+    {
+      return PlaceRelation::Disjoint;
+    }
 
     if projections_match {
       if is_sub_part {
@@ -267,10 +398,7 @@ pub fn pointer_for_place(place: Place<'tcx>, tcx: TyCtxt<'tcx>) -> Option<Place<
   place
     .iter_projections()
     .rev()
-    .find(|(_, elem)| match elem {
-      ProjectionElem::Deref => true,
-      _ => false,
-    })
+    .find(|(_, elem)| matches!(elem, ProjectionElem::Deref))
     .map(|(place_ref, _)| Place {
       local: place_ref.local,
       projection: tcx.intern_place_elems(place_ref.projection),
@@ -280,7 +408,7 @@ pub fn pointer_for_place(place: Place<'tcx>, tcx: TyCtxt<'tcx>) -> Option<Place<
 struct FindSpannedPlaces<'a, 'tcx> {
   body: &'a Body<'tcx>,
   span: Span,
-  places: Vec<(Place<'tcx>, Location)>,
+  places: HashSet<(Place<'tcx>, Location)>,
 }
 
 impl Visitor<'tcx> for FindSpannedPlaces<'_, 'tcx> {
@@ -292,16 +420,28 @@ impl Visitor<'tcx> for FindSpannedPlaces<'_, 'tcx> {
       }
     }
 
-    // Ignore debug-info places whose locations are all assigned to the start of the entry block
-    if !context.is_use() {
-      return;
-    }
+    let span = match context {
+      PlaceContext::MutatingUse(..) | PlaceContext::NonMutatingUse(..) => {
+        let source_info = self.body.source_info(location);
+        Some(source_info.span)
+      }
+      PlaceContext::NonUse(non_use) => match non_use {
+        NonUseContext::VarDebugInfo => {
+          if self.body.args_iter().any(|local| local == place.local) {
+            let source_info = self.body.local_decls()[place.local].source_info;
+            Some(source_info.span)
+          } else {
+            None
+          }
+        }
+        _ => None,
+      },
+    };
 
-    let source_info = self.body.source_info(location);
-    let span = source_info.span;
-
-    if self.span.contains(span) || span.contains(self.span) {
-      self.places.push((*place, location))
+    if let Some(span) = span {
+      if self.span.contains(span) || span.contains(self.span) {
+        self.places.insert((*place, location));
+      }
     }
   }
 }
@@ -310,10 +450,10 @@ pub fn span_to_places(body: &Body<'tcx>, span: Span) -> Vec<(Place<'tcx>, Locati
   let mut visitor = FindSpannedPlaces {
     body,
     span,
-    places: Vec::new(),
+    places: HashSet::default(),
   };
   visitor.visit_body(body);
-  visitor.places
+  visitor.places.into_iter().collect::<Vec<_>>()
 }
 
 pub fn dump_results<'tcx, A>(
@@ -325,7 +465,7 @@ where
   A: Analysis<'tcx>,
   A::Domain: DebugWithContext<A>,
 {
-  let graphviz = graphviz::Formatter::new(body, &results, graphviz::OutputStyle::AfterOnly);
+  let graphviz = graphviz::Formatter::new(body, results, graphviz::OutputStyle::AfterOnly);
   let mut buf = Vec::new();
   dot::render(&graphviz, &mut buf)?;
   let mut file = File::create("/tmp/graph.dot")?;
@@ -344,3 +484,77 @@ pub fn mir_to_string(tcx: TyCtxt<'tcx>, body: &Body<'tcx>) -> Result<String> {
   write_mir_fn(tcx, body, &mut |_, _| Ok(()), &mut buffer)?;
   Ok(String::from_utf8(buffer)?)
 }
+
+pub fn location_to_string(location: Location, body: &Body<'_>) -> String {
+  let block = &body.basic_blocks()[location.block];
+  if location.statement_index == block.statements.len() {
+    format!("{:?}", block.terminator().kind)
+  } else {
+    format!("{:?}", block.statements[location.statement_index].kind)
+  }
+}
+
+struct SimplifyMir;
+impl MirPass<'tcx> for SimplifyMir {
+  fn run_pass(&self, _tcx: TyCtxt<'tcx>, body: &mut Body<'tcx>) {
+    for block in body.basic_blocks_mut() {
+      block.retain_statements(|stmt| {
+        !matches!(
+          stmt.kind,
+          // TODO: variable_select_lhs test fails if we remove FakeRead
+          // StatementKind::FakeRead(..)
+          StatementKind::StorageLive(..) | StatementKind::StorageDead(..)
+        )
+      });
+
+      let terminator = block.terminator_mut();
+      terminator.kind = match terminator.kind {
+        TerminatorKind::FalseEdge { real_target, .. } => TerminatorKind::Goto {
+          target: real_target,
+        },
+        TerminatorKind::FalseUnwind { real_target, .. } => TerminatorKind::Goto {
+          target: real_target,
+        },
+        _ => continue,
+      }
+    }
+  }
+}
+
+pub fn get_body_with_borrowck_facts(
+  tcx: TyCtxt<'tcx>,
+  body_id: BodyId,
+) -> BodyWithBorrowckFacts<'tcx> {
+  let local_def_id = tcx.hir().body_owner_def_id(body_id);
+  let mut body_with_facts = rustc_borrowck::consumers::get_body_with_borrowck_facts(
+    tcx,
+    WithOptConstParam::unknown(local_def_id),
+  );
+
+  let body = &mut body_with_facts.body;
+  SimplifyMir.run_pass(tcx, body);
+  // simplify::SimplifyCfg::new("flowistry").run_pass(tcx, body);
+
+  body_with_facts
+}
+
+pub fn split_deref(
+  place: Place<'tcx>,
+  tcx: TyCtxt<'tcx>,
+) -> Option<(Place<'tcx>, &'tcx [PlaceElem<'tcx>])> {
+  let deref_index = place
+    .projection
+    .iter()
+    .enumerate()
+    .rev()
+    .find(|(_, elem)| matches!(elem, ProjectionElem::Deref))
+    .map(|(i, _)| i)?;
+
+  Some((
+    Place {
+      local: place.local,
+      projection: tcx.intern_place_elems(&place.projection[..deref_index]),
+    },
+    &place.projection[deref_index + 1..],
+  ))
+}
